# -*- coding: UTF-8 -*-
"""
dtaidistance.dtw
~~~~~~~~~~~~~~~~

Dynamic Time Warping (DTW)

:author: Wannes Meert
:copyright: Copyright 2017-2020 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging
import math
import array

from . import ed
from . import util
from . import util_numpy
from .util import SeriesContainer
from .exceptions import NumpyException


logger = logging.getLogger("be.kuleuven.dtai.distance")


dtw_cc = None
try:
    from . import dtw_cc
except ImportError:
    logger.debug('DTAIDistance C library not available')
    dtw_cc = None

dtw_cc_omp = None
try:
    from . import dtw_cc_omp
except ImportError:
    logger.debug('DTAIDistance C-OMP library not available')
    dtw_cc_omp = None

dtw_cc_numpy = None
try:
    from . import dtw_cc_numpy
except ImportError:
    logger.debug('DTAIDistance C-Numpy library not available')
    dtw_cc_numpy = None

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None

try:
    if util_numpy.test_without_numpy():
        raise ImportError()
    import numpy as np
    DTYPE = np.double
    argmin = np.argmin
    array_min = np.min
    array_max = np.max
except ImportError:
    np = None
    argmin = util.argmin
    array_min = min
    array_max = max


def try_import_c():
    global dtw_cc
    try:
        from . import dtw_cc
    except ImportError as exc:
        print('Cannot import C library')
        print(exc)
        dtw_cc = None


inf = float("inf")


def _check_library(include_omp=False, raise_exception=True):
    if dtw_cc is None:
        msg = "The compiled dtaidistance C library is not available.\n" + \
              "See the documentation for alternative installation options."
        logger.error(msg)
        if raise_exception:
            raise Exception(msg)
    if include_omp and dtw_cc_omp is None:
        msg = "The compiled dtaidistance C-OMP library is not available.\n" + \
              "See the documentation for alternative installation options."
        logger.error(msg)
        if raise_exception:
            raise Exception(msg)


def lb_keogh(s1, s2, window=None, max_dist=None,
             max_step=None, max_length_diff=None):
    """Lowerbound LB_KEOGH"""
    # TODO: This implementation slower than distance() in C
    if window is None:
        window = max(len(s1), len(s2))

    t = 0
    for i in range(len(s1)):
        imin = max(0, i - max(0, len(s1) - len(s2)) - window + 1)
        imax = min(len(s2), i + max(0, len(s2) - len(s1)) + window)
        ui = array_max(s2[imin:imax])
        li = array_min(s2[imin:imax])
        ci = s1[i]
        if ci > ui:
            t += abs(ci - ui)
        elif ci < li:
            t += abs(ci - li)
        else:
            pass
    return t


def ub_euclidean(s1, s2):
    """ See ed.euclidean_distance"""
    return ed.distance(s1, s2)


def distance(s1, s2,
             window=None, max_dist=None, max_step=None,
             max_length_diff=None, penalty=None, psi=None,
             use_c=False, use_pruning=False, only_ub=False):
    """
    Dynamic Time Warping.

    This function keeps a compact matrix, not the full warping paths matrix.

    :param s1: First sequence
    :param s2: Second sequence
    :param window: Only allow for maximal shifts from the two diagonals smaller than this number.
        It includes the diagonal, meaning that an Euclidean distance is obtained by setting weight=1.
    :param max_dist: Stop if the returned values will be larger than this value
    :param max_step: Do not allow steps larger than this value
    :param max_length_diff: Return infinity if length of two series is larger
    :param penalty: Penalty to add if compression or expansion is applied
    :param psi: Psi relaxation parameter (ignore start and end of matching).
        Useful for cyclical series.
    :param use_c: Use fast pure c compiled functions
    :param use_pruning: Prune values based on Euclidean distance.
        This is the same as passing ub_euclidean() to max_dist
    :param only_ub: Only compute the upper bound (Euclidean).

    Returns: DTW distance
    """
    if use_c:
        if dtw_cc is None:
            logger.warning("C-library not available, using the Python version")
        else:
            return distance_fast(s1, s2, window,
                                 max_dist=max_dist,
                                 max_step=max_step,
                                 max_length_diff=max_length_diff,
                                 penalty=penalty,
                                 psi=psi,
                                 use_pruning=use_pruning,
                                 only_ub=only_ub)
    r, c = len(s1), len(s2)
    if max_length_diff is not None and abs(r - c) > max_length_diff:
        return inf
    if window is None:
        window = max(r, c)
    if not max_step:
        max_step = inf
    else:
        max_step *= max_step
    if use_pruning or only_ub:
        max_dist = ub_euclidean(s1, s2)**2
        if only_ub:
            return max_dist
    elif not max_dist:
        max_dist = inf
    else:
        max_dist *= max_dist
    if not penalty:
        penalty = 0
    else:
        penalty *= penalty
    if psi is None:
        psi = 0
    length = min(c + 1, abs(r - c) + 2 * (window - 1) + 1 + 1 + 1)
    # print("length (py) = {}".format(length))
    dtw = array.array('d', [inf] * (2 * length))
    sc = 0
    ec = 0
    ec_next = 0
    smaller_found = False
    for i in range(psi + 1):
        dtw[i] = 0
    skip = 0
    i0 = 1
    i1 = 0
    psi_shortest = inf
    for i in range(r):
        # print("i={}".format(i))
        # print(dtw)
        skipp = skip
        skip = max(0, i - max(0, r - c) - window + 1)
        i0 = 1 - i0
        i1 = 1 - i1
        for ii in range(i1*length, i1*length+length):
            dtw[ii] = inf
        j_start = max(0, i - max(0, r - c) - window + 1)
        j_end = min(c, i + max(0, c - r) + window)
        if sc > j_start:
            j_start = sc
        smaller_found = False
        ec_next = i
        if length == c + 1:
            skip = 0
        if psi != 0 and j_start == 0 and i < psi:
            dtw[i1 * length] = 0
        for j in range(j_start, j_end):
            d = (s1[i] - s2[j])**2
            if d > max_step:
                continue
            assert j + 1 - skip >= 0
            assert j - skipp >= 0
            assert j + 1 - skipp >= 0
            assert j - skip >= 0
            dtw[i1 * length + j + 1 - skip] = d + min(dtw[i0 * length + j - skipp],
                                                      dtw[i0 * length + j + 1 - skipp] + penalty,
                                                      dtw[i1 * length + j - skip] + penalty)
            # print('({},{}), ({},{}), ({},{})'.format(i0, j - skipp, i0, j + 1 - skipp, i1, j - skip))
            # print('{}, {}, {}'.format(dtw[i0, j - skipp], dtw[i0, j + 1 - skipp], dtw[i1, j - skip]))
            # print('i={}, j={}, d={}, skip={}, skipp={}'.format(i,j,d,skip,skipp))
            # print(dtw)
            if dtw[i1 * length + j + 1 - skip] > max_dist:
                if not smaller_found:
                    sc = j + 1
                if j >= ec:
                    break
            else:
                smaller_found = True
                ec_next = j + 1
        ec = ec_next
        if psi != 0 and j_end == len(s2) and len(s1) - 1 - i <= psi:
            psi_shortest = min(psi_shortest, dtw[i1 * length + length - 1])
    if psi == 0:
        d = math.sqrt(dtw[i1 * length + min(c, c + window - 1) - skip])
    else:
        ic = min(c, c + window - 1) - skip
        vc = dtw[i1 * length + ic - psi:i1 * length + ic + 1]
        d = min(array_min(vc), psi_shortest)
        d = math.sqrt(d)
    if max_dist and d > max_dist:
        d = inf
    return d


def distance_fast(s1, s2, window=None, max_dist=None,
                  max_step=None, max_length_diff=None, penalty=None, psi=None, use_pruning=False, only_ub=False):
    """Fast C version of :meth:`distance`.

    Note: the series are expected to be arrays of the type ``double``.
    Thus ``numpy.array([1,2,3], dtype=numpy.double)`` or
    ``array.array('d', [1,2,3])``
    """
    _check_library(raise_exception=True)
    # Check that Numpy arrays for C contiguous
    s1 = util_numpy.verify_np_array(s1)
    s2 = util_numpy.verify_np_array(s2)
    # Move data to C library
    d = dtw_cc.distance(s1, s2,
                        window=window,
                        max_dist=max_dist,
                        max_step=max_step,
                        max_length_diff=max_length_diff,
                        penalty=penalty,
                        psi=psi,
                        use_pruning=use_pruning,
                        only_ub=only_ub)
    return d


def _distance_with_params(t):
    return distance(t[0], t[1], **t[2])


def _distance_c_with_params(t):
    return dtw_cc.distance(t[0], t[1], **t[2])


def warping_paths(s1, s2, window=None, max_dist=None,
                  max_step=None, max_length_diff=None, penalty=None, psi=None):
    """
    Dynamic Time Warping.

    The full matrix of all warping paths is build.

    :param s1: First sequence
    :param s2: Second sequence
    :param window: see :meth:`distance`
    :param max_dist: see :meth:`distance`
    :param max_step: see :meth:`distance`
    :param max_length_diff: see :meth:`distance`
    :param penalty: see :meth:`distance`
    :param psi: see :meth:`distance`
    :returns: (DTW distance, DTW matrix)
    """
    if np is None:
        raise NumpyException("Numpy is required for the warping_paths method")
    r, c = len(s1), len(s2)
    if max_length_diff is not None and abs(r - c) > max_length_diff:
        return inf
    if window is None:
        window = max(r, c)
    if not max_step:
        max_step = inf
    else:
        max_step *= max_step
    if not max_dist:
        max_dist = inf
    else:
        max_dist *= max_dist
    if not penalty:
        penalty = 0
    else:
        penalty *= penalty
    if psi is None:
        psi = 0
    dtw = np.full((r + 1, c + 1), inf)
    # dtw[0, 0] = 0
    for i in range(psi + 1):
        dtw[0, i] = 0
        dtw[i, 0] = 0
    i0 = 1
    i1 = 0
    sc = 0
    ec = 0
    smaller_found = False
    ec_next = 0
    for i in range(r):
        i0 = i
        i1 = i + 1
        j_start = max(0, i - max(0, r - c) - window + 1)
        j_end = min(c, i + max(0, c - r) + window)
        if sc > j_start:
            j_start = sc
        smaller_found = False
        ec_next = i
        # print('i =', i, 'skip =',skip, 'skipp =', skipp)
        # jmin = max(0, i - max(0, r - c) - window + 1)
        # jmax = min(c, i + max(0, c - r) + window)
        # print(i,jmin,jmax)
        # x = dtw[i, jmin-skipp:jmax-skipp]
        # y = dtw[i, jmin+1-skipp:jmax+1-skipp]
        # print(x,y,dtw[i+1, jmin+1-skip:jmax+1-skip])
        # dtw[i+1, jmin+1-skip:jmax+1-skip] = np.minimum(x,
        #                                                y)
        for j in range(j_start, j_end):
            # print('j =', j, 'max=',min(c, c - r + i + window))
            d = (s1[i] - s2[j])**2
            if max_step is not None and d > max_step:
                continue
            # print(i, j + 1 - skip, j - skipp, j + 1 - skipp, j - skip)
            dtw[i1, j + 1] = d + min(dtw[i0, j],
                                     dtw[i0, j + 1] + penalty,
                                     dtw[i1, j] + penalty)
            # dtw[i + 1, j + 1 - skip] = d + min(dtw[i + 1, j + 1 - skip], dtw[i + 1, j - skip])
            if dtw[i1, j + 1] > max_dist:
                if not smaller_found:
                    sc = j + 1
                if j >= ec:
                    break
            else:
                smaller_found = True
                ec_next = j + 1
    dtw = np.sqrt(dtw)
    if psi == 0:
        d = dtw[i1, min(c, c + window - 1)]
    else:
        ir = i1
        ic = min(c, c + window - 1)
        vr = dtw[ir-psi:ir+1, ic]
        vc = dtw[ir, ic-psi:ic+1]
        mir = argmin(vr)
        mic = argmin(vc)
        if vr[mir] < vc[mic]:
            dtw[ir-psi+mir+1:ir+1, ic] = -1
            d = vr[mir]
        else:
            dtw[ir, ic - psi + mic + 1:ic+1] = -1
            d = vc[mic]
    if max_dist and d > max_dist:
        d = inf
    return d, dtw


def warping_paths_fast(s1, s2, window=None, max_dist=None,
                       max_step=None, max_length_diff=None, penalty=None, psi=None):
    """Fast C version of :meth:`warping_paths`."""
    s1 = util_numpy.verify_np_array(s1)
    s2 = util_numpy.verify_np_array(s2)
    r = len(s1)
    c = len(s2)
    _check_library(raise_exception=True)
    if window is None:
        window = 0
    if max_dist is None:
        max_dist = 0
    if max_step is None:
        max_step = 0
    if max_length_diff is None:
        max_length_diff = 0
    if penalty is None:
        penalty = 0
    if psi is None:
        psi = 0
    dtw = np.full((r + 1, c + 1), inf)
    d = dtw_cc.warping_paths(dtw, s1, s2,
                             window=window,
                             max_dist=max_dist,
                             max_step=max_step,
                             max_length_diff=max_length_diff,
                             penalty=penalty,
                             psi=psi)
    return d, dtw


def distance_matrix_func(use_c=False, parallel=False, show_progress=False):
    def distance_matrix_wrapper(seqs, **kwargs):
        return distance_matrix(seqs, parallel=parallel, use_c=use_c,
                               show_progress=show_progress, **kwargs)
    return distance_matrix_wrapper


def distance_matrix(s, max_dist=None, use_pruning=False, max_length_diff=None,
                    window=None, max_step=None, penalty=None, psi=None,
                    block=None, compact=False, parallel=False,
                    use_c=False, use_mp=False, show_progress=False):
    """Distance matrix for all sequences in s.

    :param s: Iterable of series
    :param window: see :meth:`distance`
    :param max_dist: see :meth:`distance`
    :param max_step: see :meth:`distance`
    :param max_length_diff: see :meth:`distance`
    :param penalty: see :meth:`distance`
    :param psi: see :meth:`distance`
    :param block: Only compute block in matrix. Expects tuple with begin and end, e.g. ((0,10),(20,25)) will
        only compare rows 0:10 with rows 20:25.
    :param compact: Return the distance matrix as an array representing the upper triangular matrix.
    :param parallel: Use parallel operations
    :param use_c: Use c compiled Python functions
    :param use_mp: Use Multiprocessing for parallel operations (not OpenMP)
    :param show_progress: Show progress using the tqdm library. This is only supported for
        the pure Python version (thus not the C-based implementations).
    :returns: The distance matrix or the condensed distance matrix if the compact argument is true
    """
    # Check whether multiprocessing is available
    if use_c:
        _check_library(raise_exception=True)
    if use_c and parallel:
        if dtw_cc_omp is None:
            logger.warning('OMP extension not loaded, using multiprocessing')
    if parallel and (use_mp or not use_c or dtw_cc_omp is None):
        try:
            import multiprocessing as mp
            logger.info('Using multiprocessing')
        except ImportError:
            msg = 'Cannot load multiprocessing'
            logger.error(msg)
            raise Exception(msg)
    else:
        mp = None
    # Prepare options and data to pass to distance method
    dist_opts = {
        'max_dist': max_dist,
        'max_step': max_step,
        'window': window,
        'max_length_diff': max_length_diff,
        'penalty': penalty,
        'psi': psi,
        'use_pruning': use_pruning
    }
    s = SeriesContainer.wrap(s)
    if max_length_diff is None:
        max_length_diff = inf
    dists = None
    if use_c:
        for k, v in dist_opts.items():
            if v is None:
                # None is represented as 0.0 for C
                dist_opts[k] = 0.0

    logger.info('Computing distances')
    if use_c and parallel and not use_mp and dtw_cc_omp is not None:
        logger.info("Compute distances in C (parallel=OMP)")
        dist_opts['block'] = block
        dists = dtw_cc_omp.distance_matrix(s, **dist_opts)

    elif use_c and parallel and (dtw_cc_omp is None or use_mp):
        logger.info("Compute distances in C (parallel=MP)")
        idxs = _distance_matrix_idxs(block, len(s))
        with mp.Pool() as p:
            dists = p.map(_distance_c_with_params, [(s[r], s[c], dist_opts) for c, r in zip(*idxs)])

    elif use_c and not parallel:
        logger.info("Compute distances in C (parallel=No)")
        dist_opts['block'] = block
        dists = dtw_cc.distance_matrix(s, **dist_opts)

    elif not use_c and parallel:
        logger.info("Compute distances in Python (parallel=MP)")
        idxs = _distance_matrix_idxs(block, len(s))
        with mp.Pool() as p:
            dists = p.map(_distance_with_params, [(s[r], s[c], dist_opts) for c, r in zip(*idxs)])

    elif not use_c and not parallel:
        logger.info("Compute distances in Python (parallel=No)")
        dists = distance_matrix_python(s, block=block, show_progress=show_progress,
                                       max_length_diff=max_length_diff, dist_opts=dist_opts)

    else:
        raise Exception(f'Unsupported combination of: parallel={parallel}, '
                        f'use_c={use_c}, dtw_cc_omp={dtw_cc_omp}, use_mp={use_mp}')

    exp_length = _distance_matrix_length(block, len(s))
    assert len(dists) == exp_length, "len(dists)={} != {}".format(len(dists), exp_length)
    if compact:
        return dists

    # Create full matrix and fill upper triangular matrix with distance values (or only block if specified)
    dists_matrix = distances_array_to_matrix(dists, nb_series=len(s), block=block)

    return dists_matrix


def distances_array_to_matrix(dists, nb_series, block=None):
    """Transform a condensed distances array to a full matrix representation.

    The upper triangular matrix will contain all the distances.
    """
    if np is None:
        raise NumpyException("Numpy is required for the distances_array_to_matrix method")
    dists_matrix = np.full((nb_series, nb_series), inf, dtype=DTYPE)
    idxs = _distance_matrix_idxs(block, nb_series)
    dists_matrix[idxs] = dists
    # dists_cond = np.zeros(self._size_cond(len(series)))
    # idx = 0
    # for r in range(len(series) - 1):
    #     dists_cond[idx:idx + len(series) - r - 1] = dists[r, r + 1:]
    #     idx += len(series) - r - 1
    return dists_matrix


def distance_array_index(a, b, nb_series):
    if a == b:
        raise ValueError("Distance between the same series is not available.")
    if a > b:
        a, b = b, a
    idx = 0
    for r in range(a):
        idx += nb_series - r - 1
    idx += b - a - 1
    return idx


def distance_matrix_python(s, block=None, show_progress=False, max_length_diff=None, dist_opts=None):
    if dist_opts is None:
        dist_opts = {}
    dists = array.array('d', [inf] * _distance_matrix_length(block, len(s)))
    if block is None:
        it_r = range(len(s))
    else:
        it_r = range(block[0][0], block[0][1])
    if show_progress:
        it_r = tqdm(it_r)
    idx = 0
    for r in it_r:
        if block is None:
            it_c = range(r + 1, len(s))
        else:
            it_c = range(max(r + 1, block[1][0]), min(len(s), block[1][1]))
        for c in it_c:
            if abs(len(s[r]) - len(s[c])) <= max_length_diff:
                dists[idx] = distance(s[r], s[c], **dist_opts)
            idx += 1
    return dists


def _distance_matrix_idxs(block, nb_series):
    if block is None or block == 0:
        if np is not None:
            idxs = np.triu_indices(nb_series, k=1)
            return idxs
        # Numpy is not available
        block = ((0, nb_series), (0, nb_series))
    idxsl_r = []
    idxsl_c = []
    for r in range(block[0][0], block[0][1]):
        for c in range(max(r + 1, block[1][0]), min(nb_series, block[1][1])):
            idxsl_r.append(r)
            idxsl_c.append(c)
    if np is not None:
        idxs = (np.array(idxsl_r), np.array(idxsl_c))
    else:
        idxs = (idxsl_r, idxsl_c)
    return idxs


def _distance_matrix_length(block, nb_series):
    if block is not None:
        block_rb = block[0][0]
        block_re = block[0][1]
        block_cb = block[1][0]
        block_ce = block[1][1]
        length = 0
        for ri in range(block_rb, block_re):
            if block_cb <= ri:
                if block_ce > ri:
                    length += (block_ce - ri - 1)
            else:
                if block_ce > ri:
                    length += (block_ce - block_cb)
    else:
        length = int(nb_series * (nb_series - 1) / 2)
    return length


def distance_matrix_fast(s, max_dist=None, max_length_diff=None,
                         window=None, max_step=None, penalty=None, psi=None,
                         block=None, compact=False, parallel=True):
    """Fast C version of :meth:`distance_matrix`."""
    _check_library(raise_exception=True, include_omp=parallel)
    return distance_matrix(s, max_dist=max_dist, max_length_diff=max_length_diff,
                           window=window, max_step=max_step, penalty=penalty, psi=psi,
                           block=block, compact=compact, parallel=parallel,
                           use_c=True, show_progress=False)


def warping_path(from_s, to_s, **kwargs):
    """Compute warping path between two sequences."""
    dist, paths = warping_paths(from_s, to_s, **kwargs)
    path = best_path(paths)
    return path


def warping_amount(path):
    """
        Returns the number of compressions and expansions performed to obtain the best path.
        Can be used as a metric for the amount of warping.

        :param path: path to be tested

        :returns number of compressions or expansions

    """
    n = 0
    for i in range(1, len(path)):
        if path[i - 1][0] + 1 != path[i][0] or path[i - 1][1] + 1 != path[i][1]:
            n += 1

    return n


def warping_path_penalty(s1, s2, penalty_post=0, **kwargs):
    """Dynamic Time Warping with an alternative penalty.

    This function supports two different penalties. The traditional DTW penalty `penalty`
    is used in the matrix during calculation of the warping path (see :meth:`distance`).

    The second penalty `penalty_post` measures the amount of warping. This penalty doesn't
    affect the warping path and is added to the DTW distance after the warping for every compression or expansion.

    Same options as :meth:`warping_paths`

    :param s1: First sequence
    :param s2: Second sequence
    :param penalty_post: Penalty to be added after path calculation, for compression/extension

    :returns [DTW distance, best path, DTW distance between 2 path elements, DTW matrix]
    """
    dist, paths = warping_paths(s1, s2, **kwargs)
    path = best_path(paths)

    path_stepsize = []
    for i in range(1, len(path)):
        if path[i - 1][0] + 1 != path[i][0] or path[i - 1][1] + 1 != path[i][1]:
            dist += penalty_post

        path_stepsize.append(paths[path[i][0] + 1, path[i][1] + 1] - paths[path[i - 1][0] + 1, path[i - 1][1] + 1])

    return [dist, path, path_stepsize, paths]


def warp(from_s, to_s, path=None, **kwargs):
    """Warp a function to optimally match a second function.

    :param from_s: First sequence
    :param to_s: Second sequence
    :param path: (Optional) Path to use wrap the 'from_s' sequence to the 'to_s' sequence
                If provided, this function will use it.
                If not provided, this function will calculate it using the warping_path function
    :param kwargs: Same options as :meth:`warping_paths`.
    """
    if path is None:
        path = warping_path(from_s, to_s, **kwargs)
    from_s2 = array.array('d', [0] * len(to_s))
    from_s2_cnt = array.array('i', [0] * len(to_s))
    for r_c, c_c in path:
        from_s2[c_c] += from_s[r_c]
        from_s2_cnt[c_c] += 1
    for i in range(len(to_s)):
        from_s2[i] /= from_s2_cnt[i]
    return from_s2, path


def best_path(paths):
    """Compute the optimal path from the nxm warping paths matrix."""
    i, j = int(paths.shape[0] - 1), int(paths.shape[1] - 1)
    p = []
    if paths[i, j] != -1:
        p.append((i - 1, j - 1))
    while i > 0 and j > 0:
        c = argmin([paths[i - 1, j - 1], paths[i - 1, j], paths[i, j - 1]])
        if c == 0:
            i, j = i - 1, j - 1
        elif c == 1:
            i = i - 1
        elif c == 2:
            j = j - 1
        if paths[i, j] != -1:
            p.append((i - 1, j - 1))
    p.pop()
    p.reverse()
    return p


def best_path2(paths):
    """Compute the optimal path from the nxm warping paths matrix."""
    m = paths
    path = []
    r, c = m.shape
    r -= 1
    c -= 1
    v = m[r, c]
    path.append((r - 1, c - 1))
    while r > 1 or c > 1:
        r_c, c_c = r, c
        if r >= 1 and c >= 1 and m[r - 1, c - 1] <= v:
            r_c, c_c, v = r - 1, c - 1, m[r - 1, c - 1]
        if r >= 1 and m[r - 1, c] <= v:
            r_c, c_c, v = r - 1, c, m[r - 1, c]
        if c >= 1 and m[r, c - 1] <= v:
            r_c, c_c, v = r, c - 1, m[r, c - 1]
        path.append((r_c - 1, c_c - 1))
        r, c = r_c, c_c
    path.reverse()
    return path



import numpy as np
from dtaidistance import dtw, clustering
import array
import pytest
import math

n = 1
nn = 100


# --- DISTANCE 1 ---


@pytest.mark.benchmark(group="distance1")
def test_distance1_python_compress(benchmark):
    s1 = [0, 0, 1, 2, 1, 0, 1, 0, 0]*n
    s2 = [0, 1, 2, 0, 0, 0, 0, 0, 0]*n

    def d():
        return dtw.distance(s1, s2)

    assert benchmark(d) == math.sqrt(2*n)


@pytest.mark.benchmark(group="distance1")
def test_distance1_python_matrix(benchmark):
    s1 = [0, 0, 1, 2, 1, 0, 1, 0, 0]*n
    s2 = [0, 1, 2, 0, 0, 0, 0, 0, 0]*n

    def d():
        dd, _ = dtw.warping_paths(s1, s2)
        return dd

    assert benchmark(d) == math.sqrt(2*n)


@pytest.mark.benchmark(group="distance1")
def test_distance1_c_numpy(benchmark):
    s1 = np.array([0., 0, 1, 2, 1, 0, 1, 0, 0]*n)
    s2 = np.array([0., 1, 2, 0, 0, 0, 0, 0, 0]*n)

    def d():
        return dtw.distance_fast(s1, s2)

    assert benchmark(d) == math.sqrt(2*n)


@pytest.mark.benchmark(group="distance1")
def test_distance1_c_array(benchmark):
    s1 = array.array('d', [0., 0, 1, 2, 1, 0, 1, 0, 0]*n)
    s2 = array.array('d', [0., 1, 2, 0, 0, 0, 0, 0, 0]*n)

    def d():
        return dtw.distance_fast(s1, s2)

    assert benchmark(d) == math.sqrt(2*n)


@pytest.mark.benchmark(group="distance1")
def test_distance1_c_array_prune(benchmark):
    s1 = array.array('d', [0., 0, 1, 2, 1, 0, 1, 0, 0]*n)
    s2 = array.array('d', [0., 1, 2, 0, 0, 0, 0, 0, 0]*n)

    def d():
        return dtw.distance_fast(s1, s2, use_pruning=True)

    assert benchmark(d) == math.sqrt(2*n)


# --- DISTANCE MATRIX 1 ---


@pytest.mark.benchmark(group="matrix1")
def test_distance_matrix1_serial_python(benchmark):
    s = [[0, 0, 1, 2, 1, 0, 1, 0, 0] * n,
         [0, 1, 2, 0, 0, 0, 0, 0, 0] * n,
         [1, 2, 0, 0, 0, 0, 0, 1] * n] * nn
    s = [np.array(si) for si in s]

    def d():
        return dtw.distance_matrix(s, parallel=False, use_c=False, compact=True)

    m = benchmark(d)
    assert m[0] == math.sqrt(2*n)


@pytest.mark.benchmark(group="matrix1")
def test_distance_matrix1_parallel_python(benchmark):
    s = [[0, 0, 1, 2, 1, 0, 1, 0, 0] * n,
         [0, 1, 2, 0, 0, 0, 0, 0, 0] * n,
         [1, 2, 0, 0, 0, 0, 0, 1] * n] * nn
    s = [np.array(si) for si in s]

    def d():
        return dtw.distance_matrix(s, parallel=True, use_c=False, compact=True)

    m = benchmark(d)
    assert m[0] == math.sqrt(2*n)


@pytest.mark.benchmark(group="matrix1")
def test_distance_matrix1_parallel_mp_c(benchmark):
    s = [[0., 0, 1, 2, 1, 0, 1, 0, 0] * n,
         [0., 1, 2, 0, 0, 0, 0, 0, 0] * n,
         [1., 2, 0, 0, 0, 0, 0, 1] * n] * nn
    s = [np.array(si) for si in s]

    def d():
        return dtw.distance_matrix(s, parallel=True, use_c=True, use_mp=True, compact=True)

    m = benchmark(d)
    assert m[0] == math.sqrt(2*n)


@pytest.mark.benchmark(group="matrix1")
def test_distance_matrix1_serial_c(benchmark):
    s = [[0., 0, 1, 2, 1, 0, 1, 0, 0]*n,
         [0., 1, 2, 0, 0, 0, 0, 0, 0]*n,
         [1., 2, 0, 0, 0, 0, 0, 1]*n]*nn
    s = [np.array(si) for si in s]

    def d():
        return dtw.distance_matrix(s, parallel=False, use_c=True, compact=True)

    m = benchmark(d)
    assert m[0] == math.sqrt(2*n)


@pytest.mark.benchmark(group="matrix1")
def test_distance_matrix1_serial_c_pruned(benchmark):
    s = [[0., 0, 1, 2, 1, 0, 1, 0, 0]*n,
         [0., 1, 2, 0, 0, 0, 0, 0, 0]*n,
         [1., 2, 0, 0, 0, 0, 0, 1]*n]*nn
    s = [np.array(si) for si in s]

    def d():
        return dtw.distance_matrix(s, parallel=False, use_c=True, compact=True, use_pruning=True)

    m = benchmark(d)
    assert m[0] == math.sqrt(2*n)


@pytest.mark.benchmark(group="matrix1")
def test_distance_matrix1_parallel_c(benchmark):
    s = [[0., 0, 1, 2, 1, 0, 1, 0, 0] * n,
         [0., 1, 2, 0, 0, 0, 0, 0, 0] * n,
         [1., 2, 0, 0, 0, 0, 0, 1] * n] * nn
    s = [np.array(si) for si in s]

    def d():
        return dtw.distance_matrix(s, parallel=True, use_c=True, compact=True)

    m = benchmark(d)
    assert m[0] == math.sqrt(2*n), "m[0,1]={} != {}".format(m[0, 1], math.sqrt(2*n))
    assert m[0] == pytest.approx(math.sqrt(2*n))


# --- CLUSTER MATRIX 1 ---

@pytest.mark.benchmark(group="cluster1")
def test_cluster1_hierarchical(benchmark):
    s = [[0, 0, 1, 2, 1, 0, 1, 0, 0] * n,
         [0, 1, 2, 0, 0, 0, 0, 0, 0] * n,
         [1, 2, 0, 0, 0, 0, 0, 1] * n] * nn
    s = [np.array(si) for si in s]

    def d():
        c = clustering.Hierarchical(dtw.distance_matrix_fast, {})
        return c.fit(s)

    benchmark(d)


@pytest.mark.benchmark(group="cluster1")
def test_cluster1_linkage(benchmark):
    s = [[0, 0, 1, 2, 1, 0, 1, 0, 0] * n,
         [0, 1, 2, 0, 0, 0, 0, 0, 0] * n,
         [1, 2, 0, 0, 0, 0, 0, 1] * n] * nn
    s = [np.array(si) for si in s]

    def d():
        c = clustering.LinkageTree(dtw.distance_matrix_fast, {})
        return c.fit(s)

    benchmark(d)


if __name__ == "__main__":
    # test_distance1_c_numpy(lambda x: x())
    # test_cluster1_linkage(lambda x: x())
    test_distance_matrix1_serial_python(lambda x: x())



import pytest
import sys
import os
import math
sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
from dtaidistance import dtw, util_numpy


numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")


@numpyonly
def test_numpymatrix():
    """Passing a matrix instead of a list failed because the array is now a
    view instead of the original data structure."""
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
            [0., 0, 1, 2, 1, 0, 1, 0, 0],
            [0., 1, 2, 0, 0, 0, 0, 0, 0],
            [1., 2, 0, 0, 0, 0, 0, 1, 0]])
        m = dtw.distance_matrix_fast(s)
        m2 = dtw.distance_matrix(s)
        correct = np.array([
            [np.inf, 1.41421356, 1.73205081],
            [np.inf, np.inf,     1.41421356],
            [np.inf, np.inf,     np.inf]])
        assert m[0, 1] == pytest.approx(math.sqrt(2))
        assert m2[0, 1] == pytest.approx(math.sqrt(2))
        np.testing.assert_almost_equal(correct, m, decimal=4)
        np.testing.assert_almost_equal(correct, m2, decimal=4)


@numpyonly
def test_numpymatrix_compact():
    """Passing a matrix instead of a list failed because the array is now a
    view instead of the original data structure."""
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
            [0., 0, 1, 2, 1, 0, 1, 0, 0],
            [0., 1, 2, 0, 0, 0, 0, 0, 0],
            [1., 2, 0, 0, 0, 0, 0, 1, 0]])
        m = dtw.distance_matrix_fast(s, compact=True)
        m2 = dtw.distance_matrix(s, compact=True)
        correct = np.array([1.41421356, 1.73205081, 1.41421356])
        assert m[0] == pytest.approx(math.sqrt(2))
        assert m2[0] == pytest.approx(math.sqrt(2))
        np.testing.assert_almost_equal(correct, m, decimal=4)
        np.testing.assert_almost_equal(correct, m2, decimal=4)


@numpyonly
def test_numpymatrix_transpose():
    """Passing a matrix instead of a list failed because the array is now a
    view instead of the original data structure."""
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
            [0., 0., 1.,],
            [0, 1, 2],
            [1, 2, 0],
            [2, 0, 0],
            [1, 0, 0],
            [0, 0, 0],
            [1, 0, 0],
            [0, 0, 1],
            [0, 0, 0]
        ]).T
        m = dtw.distance_matrix_fast(s)
        m2 = dtw.distance_matrix(s)
        correct = np.array([
            [np.inf, 1.41421356, 1.73205081],
            [np.inf, np.inf,     1.41421356],
            [np.inf, np.inf,     np.inf]])
        assert m[0, 1] == pytest.approx(math.sqrt(2))
        assert m2[0, 1] == pytest.approx(math.sqrt(2))
        np.testing.assert_almost_equal(correct, m, decimal=4)
        np.testing.assert_almost_equal(correct, m2, decimal=4)


# def test_negativedimensions():
#     """Failed for sizes of (62706, 104): ValueError: negative dimensions are not allowed
#     """
#     s = np.array([[ 0.,  0.,  0.,  0.,  2.,  4.,  5., 10., 11., 10., 10.,  6.,  3.,
#          2.,  1.,  1.,  1.,  1.,  1.,  0.,  1.,  1.,  1.,  1.,  0.,  2.,
#          4.,  4.,  5.,  5.,  3.,  4.,  4.,  2.,  2.,  2.,  3.,  3.,  4.,
#          4.,  4.,  4.,  2.,  2.,  3.,  3.,  7.,  6.,  4.,  4.,  8.,  8.,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN],
#        [ 1., np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN],
#        [ 3.,  4., 11., 13., 12., 10.,  3.,  1.,  5.,  3., 14., 20., 37.,
#         40., 38., 31., 18., 10.,  9., 12.,  8., 16., 15., 19., 27., 26.,
#         30., 23., 18., 15., 13., 25., 28., 25., 27., 15.,  8., 11., 13.,
#         23., 18., 12.,  6., np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN,
#         np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN]])
#
#     window = 1
#     dists = dtw.distance_matrix_fast(s, window=int(0.01 * window * s.shape[1]))
#     print(dists)
#
#     dists = dtw.distance_matrix(s, window=int(0.01 * window * s.shape[1]))
#     print(dists)
#
#     wp = dtw.warping_paths(s[0], s[2])
#     print(wp)


@numpyonly
def test_negativedimensions2():
    """
    For length 62706, the computation (len_cur * (len_cur - 1)) is a number that is too large.

    For length 62706 the length calculation for the distance matrix returned 1965958512 but should return 1965989865
    Without parenthesis len_cur * (len_cur - 1) / 2 returns -181493783
    Incorrect behaviour can be simulated using fixedint.Int32(62706)
    Problem is with dividing an uneven number to optimize the computation and require smaller number.
    This dividing the uneven number should and can be avoided.
    """
    with util_numpy.test_uses_numpy() as np:
        # s = np.full((62706, 104), 1.0)
        s = np.full((10, 104), 1.0)
        window = 1
        dists = dtw.distance_matrix_fast(s, window=int(0.01 * window * s.shape[1]))
        print(dists)


@numpyonly
def test_overflowdistance():
    with util_numpy.test_uses_numpy() as np:
        maxvalthirtytwobit = 2147483647
        s = np.array([
            [maxvalthirtytwobit, maxvalthirtytwobit, 1, 2, 1, 0, 1, 0, 0],
            [1., 2, 0, 0, 0, 0, 0, 1, 0]])
        d1 = dtw.distance(s[0], s[1], use_c=False)
        d2 = dtw.distance(s[0], s[1], use_c=True)
        print(d1)
        print(d2)
        # m = dtw_c.distance_matrix_nogil(s)
        # m2 = dtw.distance_matrix(s, compact=True)
        # print(m)
        # print(m2)
        # correct = np.array([1.41421356, 1.73205081, 1.41421356])
        # assert m[0] == pytest.approx(math.sqrt(2))
        # assert m2[0] == pytest.approx(math.sqrt(2))
        # np.testing.assert_almost_equal(correct, m, decimal=4)
        # np.testing.assert_almost_equal(correct, m2, decimal=4)


if __name__ == "__main__":
    # test_numpymatrix()
    # test_numpymatrix_transpose()
    # test_negativedimensions()
    # test_negativedimensions2()
    test_overflowdistance()





from dtaidistancec_globals cimport seq_t


cdef extern from "lib/DTAIDistanceC/DTAIDistanceC/ed.h":
    seq_t euclidean_distance(seq_t *s1, size_t l1, seq_t *s2, size_t l2)
    seq_t euclidean_distance_ndim(seq_t *s1, size_t l1, seq_t *s2, size_t l2, int ndim)




from dtaidistancec_globals cimport seq_t


cdef extern from "lib/DTAIDistanceC/DTAIDistanceC/dtw.h":
    ctypedef struct DTWSettings:
        size_t window
        seq_t max_dist
        seq_t max_step
        size_t max_length_diff
        seq_t penalty
        size_t psi
        bint use_ssize_t
        bint use_pruning
        bint only_ub

    ctypedef struct DTWBlock:
        size_t rb
        size_t re
        size_t cb
        size_t ce

    DTWSettings dtw_settings_default()
    void dtw_print_settings(DTWSettings *settings)

    seq_t dtw_distance(seq_t *s1, size_t l1, seq_t *s2, size_t l2,
                          DTWSettings *settings)
    seq_t dtw_distance_ndim(seq_t *s1, size_t l1, seq_t *s2, size_t l2, int ndim,
                               DTWSettings *settings)
    seq_t dtw_warping_paths(seq_t *wps, seq_t *s1, size_t l1, seq_t *s2, int l2,
                               bint return_dtw, bint do_sqrt, DTWSettings *settings)

    seq_t ub_euclidean(seq_t *s1, size_t l1, seq_t *s2, size_t l2)
    seq_t ub_euclidean_ndim(seq_t *s1, size_t l1, seq_t *s2, size_t l2, int ndim)
    seq_t lb_keogh(seq_t *s1, size_t l1, seq_t *s2, size_t l2, DTWSettings *settings)

    DTWBlock dtw_empty_block()
    size_t dtw_distances_ptrs(seq_t **ptrs, size_t nb_ptrs, size_t* lengths, seq_t* output,
                          DTWBlock* block, DTWSettings* settings)
    size_t dtw_distances_ndim_ptrs(seq_t **ptrs, size_t nb_ptrs, size_t* lengths, int ndim, seq_t* output,
                                   DTWBlock* block, DTWSettings* settings)
    size_t dtw_distances_matrix(seq_t *matrix, size_t nb_rows, size_t nb_cols, seq_t* output,
                            DTWBlock* block, DTWSettings* settings)
    size_t dtw_distances_ndim_matrix(seq_t *matrix, size_t nb_rows, size_t nb_cols, int ndim, seq_t* output,
                                     DTWBlock* block, DTWSettings* settings)
    size_t dtw_distances_length(DTWBlock *block, size_t nb_series, bint use_ssize_t)



import os
import sys
import tempfile
import pytest
import logging
from pathlib import Path
from dtaidistance import dtw, clustering, util_numpy


logger = logging.getLogger("be.kuleuven.dtai.distance")
directory = None
numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")


@numpyonly
def test_clustering():
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
             [0., 0, 1, 2, 1, 0, 1, 0, 0],
             [0., 1, 2, 0, 0, 0, 0, 0, 0],
             [1., 2, 0, 0, 0, 0, 0, 1, 1],
             [0., 0, 1, 2, 1, 0, 1, 0, 0],
             [0., 1, 2, 0, 0, 0, 0, 0, 0],
             [1., 2, 0, 0, 0, 0, 0, 1, 1]])

        def test_hook(from_idx, to_idx, distance):
            assert (from_idx, to_idx) in [(3, 0), (4, 1), (5, 2), (1, 0)]
        model = clustering.Hierarchical(dtw.distance_matrix_fast, {}, 2, merge_hook=test_hook,
                                        show_progress=False)
        cluster_idx = model.fit(s)
        assert cluster_idx[0] == {0, 1, 3, 4}
        assert cluster_idx[2] == {2, 5}


@numpyonly
def test_clustering_tree():
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
             [0., 0, 1, 2, 1, 0, 1, 0, 0],
             [0., 1, 2, 0, 0, 0, 0, 0, 0],
             [1., 2, 0, 0, 0, 0, 0, 1, 1],
             [0., 0, 1, 2, 1, 0, 1, 0, 0],
             [0., 1, 2, 0, 0, 0, 0, 0, 0],
             [1., 2, 0, 0, 0, 0, 0, 1, 1],
             [1., 2, 0, 0, 0, 0, 0, 1, 1]])

        def test_hook(from_idx, to_idx, distance):
            assert (from_idx, to_idx) in [(3, 0), (4, 1), (5, 2), (6, 2), (1, 0), (2, 0)]
        model = clustering.Hierarchical(dtw.distance_matrix_fast, {}, merge_hook=test_hook,
                                        show_progress=False)
        modelw = clustering.HierarchicalTree(model)
        cluster_idx = modelw.fit(s)
        assert cluster_idx[0] == {0, 1, 2, 3, 4, 5, 6}

        if directory:
            hierarchy_fn = os.path.join(directory, "hierarchy.png")
            graphviz_fn = os.path.join(directory, "hierarchy.dot")
        else:
            file = tempfile.NamedTemporaryFile()
            hierarchy_fn = file.name + "_hierarchy.png"
            graphviz_fn = file.name + "_hierarchy.dot"
        modelw.plot(hierarchy_fn)
        print("Figure saved to", hierarchy_fn)
        with open(graphviz_fn, "w") as ofile:
            print(modelw.to_dot(), file=ofile)
        print("Dot saved to", graphviz_fn)


@numpyonly
def test_clustering_tree_maxdist():
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
             [0., 0, 1, 2, 1, 0, 1, 0, 0],
             [0., 1, 2, 0, 0, 0, 0, 0, 0],
             [1., 2, 0, 0, 0, 0, 0, 1, 1],
             [0., 0, 1, 2, 1, 0, 1, 0, 0],
             [0., 1, 2, 0, 0, 0, 0, 0, 0],
             [1., 2, 0, 0, 0, 0, 0, 1, 1],
             [1., 2, 0, 0, 0, 0, 0, 1, 1]])

        def test_hook(from_idx, to_idx, distance):
            assert (from_idx, to_idx) in [(3, 0), (4, 1), (5, 2), (6, 2), (1, 0), (2, 0)]
        model = clustering.Hierarchical(dtw.distance_matrix_fast, {}, merge_hook=test_hook,
                                        show_progress=False, max_dist=0.1)
        modelw = clustering.HierarchicalTree(model)
        cluster_idx = modelw.fit(s)
        assert cluster_idx[0] == {0, 1, 2, 3, 4, 5, 6}

        if directory:
            hierarchy_fn = os.path.join(directory, "hierarchy.png")
            graphviz_fn = os.path.join(directory, "hierarchy.dot")
        else:
            file = tempfile.NamedTemporaryFile()
            hierarchy_fn = file.name + "_hierarchy.png"
            graphviz_fn = file.name + "_hierarchy.dot"
        modelw.plot(hierarchy_fn)
        print("Figure saved to", hierarchy_fn)
        with open(graphviz_fn, "w") as ofile:
            print(modelw.to_dot(), file=ofile)
        print("Dot saved to", graphviz_fn)


@numpyonly
def test_linkage_tree():
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
             [0., 0, 1, 2, 1, 0, 1, 0, 0],
             [0., 1, 2, 0, 0, 0, 0, 0, 0],
             [1., 2, 0, 0, 0, 0, 0, 1, 1],
             [0., 0, 1, 2, 1, 0, 1, 0, 0],
             [0., 1, 2, 0, 0, 0, 0, 0, 0],
             [1., 2, 0, 0, 0, 0, 0, 1, 1],
             [1., 2, 0, 0, 0, 0, 0, 1, 1]])

        model = clustering.LinkageTree(dtw.distance_matrix_fast, {})
        cluster_idx = model.fit(s)

        if directory:
            hierarchy_fn = os.path.join(directory, "hierarchy.png")
            graphviz_fn = os.path.join(directory, "hierarchy.dot")
        else:
            file = tempfile.NamedTemporaryFile()
            hierarchy_fn = file.name + "_hierarchy.png"
            graphviz_fn = file.name + "_hierarchy.dot"
        model.plot(hierarchy_fn)
        print("Figure saved to", hierarchy_fn)
        with open(graphviz_fn, "w") as ofile:
            print(model.to_dot(), file=ofile)
        print("Dot saved to", graphviz_fn)


@numpyonly
def test_controlchart():
    with util_numpy.test_uses_numpy() as np:
        import matplotlib.pyplot as plt
        from matplotlib.colors import ListedColormap
        series = np.zeros((600, 60))
        rsrc_fn = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rsrc', 'synthetic_control.data')
        with open(rsrc_fn, 'r') as ifile:
            for idx, line in enumerate(ifile.readlines()):
                series[idx, :] = line.split()
        s = []
        for idx in range(0, 600, 20):
            s.append(series[idx, :])

        model = clustering.LinkageTree(dtw.distance_matrix_fast, {'parallel': True})
        cluster_idx = model.fit(s)

        if directory:
            hierarchy_fn = os.path.join(directory, "hierarchy.png")
        else:
            file = tempfile.NamedTemporaryFile()
            hierarchy_fn = file.name + "_hierarchy.png"
        fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(10, 10))
        show_ts_label = lambda idx: "ts-" + str(idx)
        # show_ts_label = list(range(len(s)))

        def curcmap(idx):
            if idx % 2 == 0:
                return 'r'
            return 'g'

        model.plot(hierarchy_fn, axes=ax, show_ts_label=show_ts_label,
                   show_tr_label=True, ts_label_margin=-10,
                   ts_left_margin=10, ts_sample_length=1, ts_color=curcmap)
        print("Figure saved to", hierarchy_fn)


@numpyonly
def test_plotbug1():
    with util_numpy.test_uses_numpy() as np:
        s1 = np.array([0., 0, 1, 2, 1, 0, 1, 0, 0, 2, 1, 0, 0])
        s2 = np.array([0., 1, 2, 3, 1, 0, 0, 0, 2, 1, 0, 0])

        series = s1, s2

        m = clustering.LinkageTree(dtw.distance_matrix, {})
        m.fit(series)

        if directory:
            hierarchy_fn = os.path.join(directory, "clustering.png")
        else:
            file = tempfile.NamedTemporaryFile()
            hierarchy_fn = file.name + "_clustering.png"
        m.plot(hierarchy_fn)
        print("Figure save to", hierarchy_fn)


if __name__ == "__main__":
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    # test_clustering_tree()
    test_clustering_tree_maxdist()
    # test_linkage_tree()
    # test_controlchart()
    # test_plotbug1()



from contextlib import ContextDecorator
import os
import logging

from .exceptions import NumpyException


logger = logging.getLogger("be.kuleuven.dtai.distance")


try:
    import numpy as np
except ImportError:
    np = None


def test_without_numpy():
    if "DTAIDISTANCE_TESTWITHOUTNUMPY" in os.environ and os.environ["DTAIDISTANCE_TESTWITHOUTNUMPY"] == "1":
        return True
    return False


def verify_np_array(seq):
    if np is not None:
        if isinstance(seq, (np.ndarray, np.generic)):
            if not seq.data.c_contiguous:
                logger.debug("Warning: Sequence 1 passed to method distance is not C-contiguous. " +
                             "The sequence will be copied.")
                seq = seq.copy(order='C')
    return seq


class NumpyStub:
    def __init__(self, testwithoutnp):
        self.testwithoutnp = testwithoutnp

    def __getattr__(self, name):
        if self.testwithoutnp or np is None:
            raise NumpyException("Numpy excepted to be available for test. "
                                 "Set DTAIDISTANCE_TESTWITHOUTNUMPY=1 to test without Numpy.")
        return getattr(np, name)


class test_uses_numpy(ContextDecorator):
    def __init__(self, strict=True):
        """Context to construct tests that use the optional dependency Numpy.

        :param strict: Throw error if Numpy is not used (to remove context where not necessary)
        :return: Numpy stub
        """
        self.strict = strict
        self.testwithoutnp = test_without_numpy()

    def __enter__(self):
        return NumpyStub(self.testwithoutnp)

    def __exit__(self, *exc):
        if self.testwithoutnp:
            if exc[0] is None:
                if self.strict and self.testwithoutnp:
                    # If no NumpyException is thrown, this test did not use Numpy because no error was thrown
                    # and should not use this decorator
                    raise Exception("Test does not use Numpy, remove decorator!")
                else:
                    return
            if issubclass(exc[0], NumpyException):
                return True



"""
dtaidistance.dtw_cc_omp
~~~~~~~~~~~~~~~~~~~~~~~

Dynamic Time Warping (DTW), C implementation, with OpenMP support.

:author: Wannes Meert
:copyright: Copyright 2020 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
from cpython cimport array
import array
from dtw_cc cimport DTWSeriesMatrix, DTWSeriesMatrixNDim, DTWSeriesPointers, DTWSettings, DTWBlock
from dtw_cc import dtw_series_from_data, distance_matrix_length
cimport dtaidistancec_dtw
cimport dtaidistancec_dtw_omp


def distance_matrix(cur, block=None, **kwargs):
    """Compute a distance matrix between all sequences given in `cur`.
    This method calls a pure c implementation of the dtw computation that
    avoids the GIL.

    Assumes C-contiguous arrays.

    :param cur: DTWSeriesMatrix or DTWSeriesPointers
    :param block: see DTWBlock
    :param kwargs: Settings (see DTWSettings)
    :return: The distance matrix as a list representing the triangular matrix.
    """
    cdef DTWSeriesMatrix matrix
    cdef DTWSeriesPointers ptrs
    cdef Py_ssize_t length = 0
    cdef Py_ssize_t block_rb=0
    cdef Py_ssize_t block_re=0
    cdef Py_ssize_t block_cb=0
    cdef Py_ssize_t block_ce=0
    cdef Py_ssize_t ri = 0
    if block is not None and block != 0.0:
        block_rb = block[0][0]
        block_re = block[0][1]
        block_cb = block[1][0]
        block_ce = block[1][1]

    settings = DTWSettings(**kwargs)
    cdef DTWBlock dtwblock = DTWBlock(rb=block_rb, re=block_re, cb=block_cb, ce=block_ce)
    length = distance_matrix_length(dtwblock, len(cur))

    # Correct block
    if dtwblock.re == 0:
        dtwblock.re_set(len(cur))
    if dtwblock.ce == 0:
        dtwblock.ce_set(len(cur))

    cdef array.array dists = array.array('d')
    array.resize(dists, length)

    if isinstance(cur, DTWSeriesMatrix) or isinstance(cur, DTWSeriesPointers):
        pass
    elif cur.__class__.__name__ == "SeriesContainer":
        cur = cur.c_data()
    else:
        cur = dtw_series_from_data(cur)

    if isinstance(cur, DTWSeriesPointers):
        ptrs = cur
        dtaidistancec_dtw_omp.dtw_distances_ptrs_parallel(
            ptrs._ptrs, ptrs._nb_ptrs, ptrs._lengths,
            dists.data.as_doubles, &dtwblock._block, &settings._settings)
    elif isinstance(cur, DTWSeriesMatrix):
        matrix = cur
        dtaidistancec_dtw_omp.dtw_distances_matrix_parallel(
            &matrix._data[0,0], matrix.nb_rows, matrix.nb_cols,
            dists.data.as_doubles, &dtwblock._block, &settings._settings)

    return dists


def distance_matrix_ndim(cur, int ndim, block=None, **kwargs):
    """Compute a distance matrix between all sequences given in `cur`.
    This method calls a pure c implementation of the dtw computation that
    avoids the GIL.

    Assumes C-contiguous arrays.

    :param cur: DTWSeriesMatrix or DTWSeriesPointers
    :param block: see DTWBlock
    :param kwargs: Settings (see DTWSettings)
    :return: The distance matrix as a list representing the triangular matrix.
    """
    cdef DTWSeriesMatrix matrix
    cdef DTWSeriesMatrixNDim matrixnd
    cdef DTWSeriesPointers ptrs
    cdef Py_ssize_t length = 0
    cdef Py_ssize_t block_rb=0
    cdef Py_ssize_t block_re=0
    cdef Py_ssize_t block_cb=0
    cdef Py_ssize_t block_ce=0
    cdef Py_ssize_t ri = 0
    if block is not None and block != 0.0:
        block_rb = block[0][0]
        block_re = block[0][1]
        block_cb = block[1][0]
        block_ce = block[1][1]

    settings = DTWSettings(**kwargs)
    cdef DTWBlock dtwblock = DTWBlock(rb=block_rb, re=block_re, cb=block_cb, ce=block_ce)
    length = distance_matrix_length(dtwblock, len(cur))

    # Correct block
    if dtwblock.re == 0:
        dtwblock.re_set(len(cur))
    if dtwblock.ce == 0:
        dtwblock.ce_set(len(cur))

    cdef array.array dists = array.array('d')
    array.resize(dists, length)

    if isinstance(cur, DTWSeriesMatrix) or isinstance(cur, DTWSeriesPointers):
        pass
    elif cur.__class__.__name__ == "SeriesContainer":
        cur = cur.c_data()
    else:
        cur = dtw_series_from_data(cur, force_pointers=True)

    if isinstance(cur, DTWSeriesPointers):
        ptrs = cur
        dtaidistancec_dtw_omp.dtw_distances_ndim_ptrs_parallel(
            ptrs._ptrs, ptrs._nb_ptrs, ptrs._lengths, ndim,
            dists.data.as_doubles, &dtwblock._block, &settings._settings)
    elif isinstance(cur, DTWSeriesMatrix):
        # This is not a n-dimensional case ?
        matrix = cur
        dtaidistancec_dtw_omp.dtw_distances_matrix_parallel(
            &matrix._data[0,0], matrix.nb_rows, matrix.nb_cols,
            dists.data.as_doubles, &dtwblock._block, &settings._settings)
    elif isinstance(cur, DTWSeriesMatrixNDim):
        matrixnd = cur
        dtaidistancec_dtw_omp.dtw_distances_ndim_matrix_parallel(
            &matrixnd._data[0,0,0], matrixnd.nb_rows, matrixnd.nb_cols, ndim,
            dists.data.as_doubles, &dtwblock._block, &settings._settings)
    else:
        raise Exception("Unknown series container")


    return dists


from __future__ import print_function
import sys
import logging
import io
import pytest
try:
    from pathlib import Path
except ImportError:
    try:
        from pathlib2 import Path  # For Python2
    except ImportError:
        raise ImportError("No pathlib or pathlib2 found")
from dtaidistance import dtw_weighted as dtww
from dtaidistance import dtw_visualisation as dtwvis
from dtaidistance import dtw, util_numpy
from dtaidistance.util import prepare_directory


logger = logging.getLogger("be.kuleuven.dtai.distance")
directory = None
numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")


def plot_series(s, l, idx=None):
    global directory
    if directory is None:
        directory = prepare_directory()
    import matplotlib.pyplot as plt
    colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
    fig1, ax1 = plt.subplots(nrows=len(s), ncols=1)
    fig2, ax2 = plt.subplots(nrows=1, ncols=1)
    for i, si in enumerate(s):
        if i == idx:
            color = colors[0]
        else:
            color = colors[int(1 + l[i])]
        ax1[i].plot(si, color=color)
        ax2.plot(si, color=color)
    fig1.savefig(str(directory / "series1.png"))
    fig2.savefig(str(directory / "series2.png"))


def plot_margins(serie, weights, clfs, importances=None):
    global directory
    if directory is None:
        directory = prepare_directory()
    try:
        from sklearn import tree
    except ImportError:
        return
    feature_names = ["f{} ({}, {})".format(i // 2, i, '-' if (i % 2) == 0 else '+') for i in range(2 * len(serie))]
    out_str = io.StringIO()
    for clf in clfs:
        tree.export_graphviz(clf, out_file=out_str, feature_names=feature_names)
        print("\n\n", file=out_str)
    with open(str(directory / "tree.dot"), "w") as ofile:
        print(out_str.getvalue(), file=ofile)
    dtww.plot_margins(serie, weights, filename=str(directory / "margins.png"), importances=importances)


@numpyonly
def test_distance1():
    with util_numpy.test_uses_numpy() as np:
        directory = prepare_directory()

        s1 = np.array([0., 0, 1, 2, 1, 0, 1, 0, 0, 2, 1, 0, 0])
        s2 = np.array([0., 1, 2, 3, 1, 10, 1, 0, 2, 1, 0, 0, 0])
        d, paths = dtw.warping_paths(s1, s2)
        # print(d, "\n", paths)
        dtwvis.plot_warpingpaths(s1, s2, paths, filename=directory / "temp1.png")

        weights = np.full((len(s1), 8), np.inf)
        weights[:, 2:4] = 0.0
        weights[4:7, 2:4] = 10.0
        weights[:, 4:6] = 0.0
        weights[4:7, 4:6] = 10.0
        d, paths = dtww.warping_paths(s1, s2, weights)
        # print(d, "\n", paths)
        dtwvis.plot_warpingpaths(s1, s2, paths, filename=directory / "temp2.png")


@numpyonly
def test_distance2():
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
            [0., 0, 1, 2, 1, 0, 1.3, 0, 0],
            [0., 0, 1, 2, 1, 0, 1,   0, 0],
            [0., 1, 2, 0, 0, 0, 0,   0, 0],
            [0., 1, 2, 0, 0, 0, 0,   0, 0],
            [1., 2, 0, 0, 0, 0, 0,   1, 1],
            [1., 2, 0, 0, 0, 0, 0,   1, 1],
            [1., 2, 0, 0, 1, 0, 0,   1, 1]])
        l = np.array([1, 1, 1, 1, 0, 0, 0])

        if directory:
            plot_series(s, l)
            savefig = str(directory / "dts.dot")
        else:
            savefig = None

        prototypeidx = 0
        ml_values, cl_values, clfs, importances = \
            dtww.series_to_dt(s, l, prototypeidx, max_clfs=50, savefig=savefig)
        logger.debug(f"ml_values = {dict(ml_values)}")
        logger.debug(f"cl_values = {dict(cl_values)}")
        weights = dtww.compute_weights_from_mlclvalues(s[prototypeidx], ml_values, cl_values, only_max=False, strict_cl=True)

        if directory:
            plot_margins(s[prototypeidx], weights, clfs)


@numpyonly
def test_distance3():
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
            [0., 0, 1, 2, 1, 0, 1.3, 0, 0],
            [0., 1, 2, 0, 0, 0, 0, 0, 0]
        ])
        w = np.array([[np.inf, np.inf, 0.,  0., 0.,  0.,   np.inf, np.inf],
                      [np.inf, np.inf, 1.1, 1., 0.,  0.,   np.inf, np.inf],
                      [np.inf, np.inf, 1.1, 1., 0.,  0.,   np.inf, np.inf],
                      [np.inf, np.inf, 0.,  0., 2.,  2.2,  np.inf, np.inf],
                      [np.inf, np.inf, 0.,  0., 1.,  1.1,  np.inf, np.inf],
                      [np.inf, np.inf, 0.,  0., 0.,  0.,   np.inf, np.inf],
                      [np.inf, np.inf, 0.,  0., 1.3, 1.43, np.inf, np.inf],
                      [np.inf, np.inf, 0.,  0., 0.,  0.,   np.inf, np.inf],
                      [np.inf, np.inf, 0.,  0., 0.,  0.,   np.inf, np.inf]])

        d, paths = dtww.warping_paths(s[0], s[1], w, window=0)
        path = dtw.best_path(paths)
        if directory:
            wp_fn = directory / "warping_paths.png"
            dtwvis.plot_warpingpaths(s[0], s[1], paths, path, filename=wp_fn)


@numpyonly
def test_distance4():
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
            [0., 0, 1,    2,    1,   0,   1.3, 0, 0],  # 0
            [0., 1, 2,    0,    0,   0,   0,   0, 0],  # 1
            [1., 2, 0,    0,    0,   0,   0,   1, 1],  # 2
            [0., 0, 1,    2,    1,   0,   1,   0, 0],  # 3
            [0., 1, 2,    0,    0,   0,   0,   0, 0],  # 4
            [1., 2, 0,    0,    0,   0,   0,   1, 1],  # 5
            [1., 2, 0,    0,    1,   0,   0,   1, 1],  # 6
            [1., 2, 0.05, 0.01, 0.9, 0,   0,   1, 1]]) # 7
        l = np.array([1, 0, 0, 1, 0, 0, 0, 0])

        if directory:
            plot_series(s, l)
            savefig = str(directory / "dts.dot")
        else:
            savefig = None

        prototypeidx = 0
        ml_values, cl_values, clf, importances = \
            dtww.series_to_dt(s, l, prototypeidx, window=2, min_ig=0.1, savefig=savefig)
        logger.debug(f"ml_values = {dict(ml_values)}")
        logger.debug(f"cl_values = {dict(cl_values)}")
        weights = dtww.compute_weights_from_mlclvalues(s[prototypeidx], ml_values, cl_values,
                                                       only_max=False, strict_cl=True)
        if directory:
            plot_margins(s[prototypeidx], weights, clf)


@numpyonly
def test_distance5():
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
            [0., 0, 0, 2,  0, -2, 0, 0,  0, 0, 0, 0,  0, 0, 0],  # 0
            [0., 0, 2, 0, -2,  0, 2, 0, -2, 0, 2, 0, -2, 0, 0],  # 1
            [0., 0, 2, 0,  0,  0, 2, 0,  0, 0, 2, 0,  0, 0, 0]   # 2
        ])
        l = np.array([1, 1, 0])

        if directory:
            plot_series(s, l)

        prototypeidx = 0
        ml_values, cl_values, clf, importances = dtww.series_to_dt(s, l, prototypeidx, window=4)
        logger.debug(f"ml_values = {dict(ml_values)}")
        logger.debug(f"cl_values = {dict(cl_values)}")
        weights = dtww.compute_weights_from_mlclvalues(s[prototypeidx], ml_values, cl_values,
                                                       only_max=False, strict_cl=True)
        if directory:
            plot_margins(s[prototypeidx], weights, clf)


@pytest.mark.skip("Takes too long")
def test_distance6():
    with util_numpy.test_uses_numpy() as np:
        s = np.loadtxt(Path(__file__).parent / "rsrc" / "series_0.csv", delimiter=',')
        l = np.loadtxt(Path(__file__).parent / "rsrc" / "labels_0.csv", delimiter=',')

        if directory:
            plot_series(s, l)
            savefig = str(directory / "dts.dot")
        else:
            savefig = None

        prototypeidx = 3
        labels = np.zeros(l.shape)
        labels[l == l[prototypeidx]] = 1
        ml_values, cl_values, clf, importances = \
            dtww.series_to_dt(s, labels, prototypeidx, window=0, min_ig=0.1, savefig=savefig)
        logger.debug(f"ml_values = {dict(ml_values)}")
        logger.debug(f"cl_values = {dict(cl_values)}")
        weights = dtww.compute_weights_from_mlclvalues(s[prototypeidx], ml_values, cl_values,
                                                       only_max=False, strict_cl=True)
        if directory:
            plot_margins(s[prototypeidx], weights, clf, prototypeidx)


@numpyonly
def test_distance7():
    with util_numpy.test_uses_numpy() as np:
        s = np.array([
            [0.0, 0.3, 0.5, 0.8, 1.0, 0.1, 0.0, 0.1],
            [0.0, 0.2, 0.3, 0.7, 1.1, 0.0, 0.1, 0.0],
            [0.1, 0.0, 1.0, 1.0, 1.0, 0.9, 0.0, 0.0],
            [0.0, 0.0, 1.1, 0.9, 1.0, 1.0, 0.0, 0.0],
            [0.0, 0.1, 1.1, 1.0, 0.9, 0.9, 0.0, 0.0],
            [0.0, 0.1, 1.0, 1.1, 0.9, 1.0, 0.0, 0.1],
            [0.0, 0.1, 0.4, 0.3, 0.2, 0.3, 0.0, 0.0],
            [0.1, 0.0, 0.2, 0.1, 0.2, 0.2, 0.1, 0.1]])
        l = np.array([1, 1, 0, 0, 0, 0, 0, 0])
        prototypeidx = 0

        if directory:
            plot_series(s, l, prototypeidx)
            savefig = str(directory / "dts.dot")
        else:
            savefig = None
        ml_values, cl_values, clf, imp = dtww.series_to_dt(s, l, prototypeidx, window=0, min_ig=0.01,
                                                           savefig=savefig,
                                                           warping_paths_fnc=dtww.warping_paths)
        # logger.debug(f"ml_values = {dict(ml_values)}")
        # logger.debug(f"cl_values = {dict(cl_values)}")
        weights = dtww.compute_weights_from_mlclvalues(s[prototypeidx], ml_values, cl_values,
                                                       only_max=False, strict_cl=True)
        if directory:
            plot_margins(s[prototypeidx], weights, clf, imp)


if __name__ == "__main__":
    # Print options
    # np.set_printoptions(precision=2)
    # Logger options
    logger.setLevel(logging.DEBUG)
    sh = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(message)s')
    sh.setFormatter(formatter)
    logger.addHandler(sh)
    logger.propagate = 0
    # Output path
    directory = Path(__file__).resolve().parent.parent / "tests" / "output"

    # Functions
    # test_distance1()
    # test_distance2()
    # test_distance3()
    # test_distance4()
    # test_distance5()
    # test_distance6()
    test_distance7()



# -*- coding: UTF-8 -*-
"""
dtaidistance.ed
~~~~~~~~~~~~~~~

Euclidean Distance (ED)

:author: Wannes Meert
:copyright: Copyright 2020 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging
import math

from . import util_numpy


logger = logging.getLogger("be.kuleuven.dtai.distance")


ed_cc = None
try:
    from . import ed_cc
except ImportError:
    logger.debug('DTAIDistance C library not available')
    ed_cc = None


try:
    if util_numpy.test_without_numpy():
        raise ImportError()
    import numpy as np
except ImportError:
    np = None


def _check_library(raise_exception=True):
    if ed_cc is None:
        msg = "The compiled dtaidistance C library is not available.\n" + \
              "See the documentation for alternative installation options."
        logger.error(msg)
        if raise_exception:
            raise Exception(msg)


def distance(s1, s2):
    """ Euclidean distance between two sequences. Supports different lengths.

    If the two series differ in length, compare the last element of the shortest series
    to the remaining elements in the longer series. This is compatible with Euclidean
    distance being used as an upper bound for DTW.

    :param s1: Sequence of numbers
    :param s2: Sequence of numbers
    :return: Euclidean distance
    """
    n = min(len(s1), len(s2))
    ub = 0
    for v1, v2 in zip(s1, s2):
        ub += (v1 - v2)**2
    # If the two series differ in length, compare the last element of the shortest series
    # to the remaining elements in the longer series
    if len(s1) > len(s2):
        v2 = s2[n - 1]
        for v1 in s1[n:]:
            ub += (v1 - v2)**2
    elif len(s1) < len(s2):
        v1 = s1[n-1]
        for v2 in s2[n:]:
            ub += (v1 - v2)**2
    return math.sqrt(ub)


def distance_fast(s1, s2):
    _check_library(raise_exception=True)
    # Check that Numpy arrays for C contiguous
    s1 = util_numpy.verify_np_array(s1)
    s2 = util_numpy.verify_np_array(s2)
    # Move data to C library
    d = ed_cc.distance(s1, s2)
    return d



"""
dtaidistance.dtw_cc_numpy
~~~~~~~~~~~~~~~~~~~~~~~~~

Dynamic Time Warping (DTW), C implementation, numpy specific extensions.

:author: Wannes Meert
:copyright: Copyright 2020 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import dtw_cc
from libc.stdint cimport intptr_t
import numpy as np
cimport numpy as np
import logging


logger = logging.getLogger("be.kuleuven.dtai.distance")


DTYPE = np.double
ctypedef np.double_t DTYPE_t


def dtw_series_from_numpy(np.ndarray[DTYPE_t, ndim=2] data):
    if not data.flags.c_contiguous:
            logger.debug("Warning: The numpy array or matrix passed to method distance_matrix is not C-contiguous. " +
                         "The array will be copied.")
            data = data.copy(order='C')
    ptrs = dtw_cc.DTWSeriesMatrix(data)
    return ptrs


def dtw_series_from_numpy_ndim(np.ndarray[DTYPE_t, ndim=3] data):
    if not data.flags.c_contiguous:
            logger.debug("Warning: The numpy array or matrix passed to method distance_matrix is not C-contiguous. " +
                         "The array will be copied.")
            data = data.copy(order='C')
    ptrs = dtw_cc.DTWSeriesMatrixNDim(data)
    return ptrs



# -*- coding: UTF-8 -*-
"""
dtaidistance.dp
~~~~~~~~~~~~~~~

Generic Dynamic Programming functions

:author: Wannes Meert
:copyright: Copyright 2017-2018 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging
import numpy as np


logger = logging.getLogger("be.kuleuven.dtai.distance")


def dp(s1, s2, fn, border=None, window=None, max_dist=None,
       max_step=None, max_length_diff=None, penalty=None, psi=None):
    """
    Generic dynamic programming.

    This function does not optimize storage when a window size is given (e.g. in contrast with
    the fast DTW functions).

    :param s1: First sequence
    :param s2: Second sequence
    :param fn: Function to compare two items from both sequences
    :param border: Callable object to fill in the initial borders (border(row_idx, col_idx).
    :param window: see :meth:`distance`
    :param max_dist: see :meth:`distance`
    :param max_step: see :meth:`distance`
    :param max_length_diff: see :meth:`distance`
    :param penalty: see :meth:`distance`
    :param psi: see :meth:`distance`
    :returns: (DTW distance, DTW matrix)
    """
    r, c = len(s1), len(s2)
    if max_length_diff is not None and abs(r - c) > max_length_diff:
        return np.inf
    if window is None:
        window = max(r, c)
    if not max_step:
        max_step = np.inf
    if not max_dist:
        max_dist = np.inf
    if not penalty:
        penalty = 0
    if psi is None:
        psi = 0
    dtw = np.full((r + 1, c + 1), np.inf)
    if border:
        for ci in range(c + 1):
            dtw[0, ci] = border(0, ci)
        for ri in range(1, r + 1):
            dtw[ri, 0] = border(ri, 0)
    for i in range(psi + 1):
        dtw[0, i] = 0
        dtw[i, 0] = 0
    last_under_max_dist = 0
    i0 = 1
    i1 = 0
    for i in range(r):
        if last_under_max_dist == -1:
            prev_last_under_max_dist = np.inf
        else:
            prev_last_under_max_dist = last_under_max_dist
        last_under_max_dist = -1
        i0 = i
        i1 = i + 1
        for j in range(max(0, i - max(0, r - c) - window + 1), min(c, i + max(0, c - r) + window)):
            d, d_indel = fn(s1[i], s2[j])
            if max_step is not None:
                if d > max_step:
                    d = np.inf
                if d_indel > max_step:
                    d_indel = np.inf
                if d > max_step and d_indel > max_step:
                    continue
            # print(f"[{i1},{j+1}] -> [{s1[i]},{s2[j]}] -> {d},{d_indel}")
            dtw[i1, j + 1] = min(d + dtw[i0, j],
                                 d_indel + dtw[i0, j + 1] + penalty,
                                 d_indel + dtw[i1, j] + penalty)
            if max_dist is not None:
                if dtw[i1, j + 1] <= max_dist:
                    last_under_max_dist = j
                else:
                    dtw[i1, j + 1] = np.inf
                    if prev_last_under_max_dist < j + 1:
                        break
        if max_dist is not None and last_under_max_dist == -1:
            return np.inf, dtw
    if psi == 0:
        d = dtw[i1, min(c, c + window - 1)]
    else:
        ir = i1
        ic = min(c, c + window - 1)
        vr = dtw[ir-psi:ir+1, ic]
        vc = dtw[ir, ic-psi:ic+1]
        mir = np.argmin(vr)
        mic = np.argmin(vc)
        if vr[mir] < vc[mic]:
            dtw[ir-psi+mir+1:ir+1, ic] = -1
            d = vr[mir]
        else:
            dtw[ir, ic - psi + mic + 1:ic+1] = -1
            d = vc[mic]
    return d, dtw




from dtaidistancec_globals cimport seq_t
from dtaidistancec_dtw cimport DTWBlock, DTWSettings


cdef extern from "lib/DTAIDistanceC/DTAIDistanceC/dtw_openmp.h":
    size_t dtw_distances_ptrs_parallel(seq_t **ptrs, size_t nb_ptrs, size_t* lengths,
                                       seq_t* output, DTWBlock* block, DTWSettings* settings)
    size_t dtw_distances_ndim_ptrs_parallel(seq_t **ptrs, size_t nb_ptrs, size_t* lengths, int ndim,
                                            seq_t* output, DTWBlock* block, DTWSettings* settings)
    size_t dtw_distances_matrix_parallel(seq_t *matrix, size_t nb_rows, size_t nb_cols,
                                         seq_t* output, DTWBlock* block, DTWSettings* settings)
    size_t dtw_distances_ndim_matrix_parallel(seq_t *matrix, size_t nb_rows, size_t nb_cols, int ndim,
                                              seq_t* output, DTWBlock* block, DTWSettings* settings)



"""
dtaidistance.ed_cc
~~~~~~~~~~~~~~~~~~

Euclidean Distance (ED), C implementation.

:author: Wannes Meert
:copyright: Copyright 2020 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging
cimport dtaidistancec_ed


logger = logging.getLogger("be.kuleuven.dtai.distance")


def distance(double[:] s1, double[:] s2):
    """ Euclidean distance between two sequences. Supports different lengths.

    If the two series differ in length, compare the last element of the shortest series
    to the remaining elements in the longer series.

    :param s1: Sequence of numbers
    :param s2: Sequence of numbers
    :return: Euclidean distance
    """
    return dtaidistancec_ed.euclidean_distance(&s1[0], len(s1), &s2[0], len(s2))


def distance_ndim(double[:, :] s1, double[:, :] s2):
    """ Euclidean distance between two sequences. Supports different lengths.

    If the two series differ in length, compare the last element of the shortest series
    to the remaining elements in the longer series.

    :param s1: Sequence of numbers
    :param s2: Sequence of numbers
    :return: Euclidean distance
    """
    # Assumes C contiguous
    if s1.shape[1] != s2.shape[1]:
        raise Exception(f"Dimension of sequence entries needs to be the same: {s1.shape[1]} != {s2.shape[1]}")
    ndim = s1.shape[1]
    return dtaidistancec_ed.euclidean_distance_ndim(&s1[0,0], len(s1), &s2[0,0], len(s2), ndim)



import math
import array

import pytest

from dtaidistance import dtw, dtw_cc, util_numpy


numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")


def test_distance1_array():
    s1 = array.array('d', [0, 0, 1, 2, 1, 0, 1, 0, 0])
    s2 = array.array('d', [0, 1, 2, 0, 0, 0, 0, 0, 0])
    d = dtw_cc.distance(s1, s2)
    # print(f'd = {d}')
    assert d == pytest.approx(math.sqrt(2))


@numpyonly
def test_distance1_numpy():
    with util_numpy.test_uses_numpy() as np:
        s1 = np.array([0., 0, 1, 2, 1, 0, 1, 0, 0])
        s2 = np.array([0., 1, 2, 0, 0, 0, 0, 0, 0])
        d = dtw.distance_fast(s1, s2)
        # print(f'd = {d}')
        assert d == pytest.approx(math.sqrt(2))


if __name__ == "__main__":
    test_distance1_array()
    # test_distance1_numpy(,)



import pytest
import os
import random
from pathlib import Path
from dtaidistance import dtw, util_numpy
from dtaidistance import dtw_visualisation as dtwvis


directory = None
numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")


@numpyonly
def test_normalize():
    with util_numpy.test_uses_numpy() as np:
        s1 = np.array([0., 0, 1, 2, 1, 0, 1, 0, 0, 2, 1, 0, 0])
        s2 = np.array([0., 1, 2, 3, 1, 0, 0, 0, 2, 1, 0, 0, 0])
        r, path = dtw.warp(s1, s2)
        if directory:
            dtwvis.plot_warp(s1, s2, r, path, filename=str(directory / "test_normalize1.png"))
        r_c = np.array([0., 1., 2., 2., 1., 0.5, 0., 0., 2., 1., 0., 0., 0.])
        if directory:
            path = dtw.warping_path(s1, s2, psi=2)
            dtwvis.plot_warping(s1, s2, path, filename=str(directory / "test_normalize2.png"))
        np.testing.assert_almost_equal(r, r_c, decimal=4)


@numpyonly
def test_normalize2():
    with util_numpy.test_uses_numpy() as np:
        s1 = np.array([0., 0, 1, 2, 1, 0, 1, 0, 0, 2, 1, 0, 0])
        s2 = np.array([0., 1, 2, 3, 1, 0, 0, 0, 2, 1, 0, 0, 0])
        d1, paths1 = dtw.warping_paths(s1, s2, psi=2)
        d2, paths2 = dtw.warping_paths_fast(s1, s2, psi=2)
        path1 = dtw.best_path(paths1)
        path2 = dtw.best_path(paths2)
        if directory:
            dtwvis.plot_warpingpaths(s1, s2, paths1, path1, filename=directory / "normalize.png")
        np.testing.assert_almost_equal(d1, d2, decimal=4)
        np.testing.assert_almost_equal(paths1, paths2, decimal=4)
        np.testing.assert_almost_equal(path1, path2, decimal=4)


@numpyonly
def test_psi_dtw_1a():
    with util_numpy.test_uses_numpy() as np:
        x = np.arange(0, 20, .5)
        s1 = np.sin(x)
        s2 = np.sin(x - 1)
        # Add noise
        # random.seed(1)
        # for idx in range(len(s2)):
        #     if random.random() < 0.05:
        #         s2[idx] += (random.random() - 0.5) / 2
        d, paths = dtw.warping_paths(s1, s2, psi=2, window=25)
        path = dtw.warping_path(s1, s2, psi=2)
        if directory:
            dtwvis.plot_warpingpaths(s1, s2, paths, path, filename=str(directory / "test_psi_dtw_1a.png"))
        # print(paths[:,:])
        # dtwvis.plot_warping(s1, s2, path, filename=os.path.expanduser("~/Desktop/test_psi_dtw_1_1.png"))
        # path = dtw.best_path(paths)
        # dtwvis.plot_warpingpaths(s1, s2, paths, path, filename=os.path.expanduser("~/Desktop/test_psi_dtw_1_2.png"))
        np.testing.assert_equal(d, 0.0)


@numpyonly
def test_psi_dtw_1b():
    with util_numpy.test_uses_numpy() as np:
        x = np.arange(0, 20, .5)
        s1 = np.sin(x)
        s2 = np.sin(x - 1)
        d = dtw.distance(s1, s2, psi=2)
        np.testing.assert_equal(d, 0.0)


@numpyonly
def test_psi_dtw_1c():
    with util_numpy.test_uses_numpy() as np:
        x = np.arange(0, 20, .5)
        s1 = np.sin(x)
        s2 = np.sin(x - 1)
        d = dtw.distance_fast(s1, s2, psi=2)
        np.testing.assert_equal(d, 0.0)


@numpyonly
def test_psi_dtw_2a():
    with util_numpy.test_uses_numpy() as np:
        x = np.arange(0, 20, .5)
        s1 = np.sin(x - 1)
        s2 = np.sin(x)
        d, paths = dtw.warping_paths(s1, s2, psi=2, window=3)
        # path = dtw.warping_path(s1, s2, psi=2)
        # dtwvis.plot_warping(s1, s2, path, filename=os.path.expanduser("~/Desktop/test_psi_dtw_2_1.png"))
        # path = dtw.best_path(paths)
        # dtwvis.plot_warpingpaths(s1, s2, paths, path, filename=os.path.expanduser("~/Desktop/test_psi_dtw_2_2.png"))
        np.testing.assert_equal(d, 0.0)


@numpyonly
def test_psi_dtw_2b():
    with util_numpy.test_uses_numpy() as np:
        x = np.arange(0, 20, .5)
        s1 = np.sin(x - 1)
        s2 = np.sin(x)
        d = dtw.distance(s1, s2, psi=2, window=3)
        np.testing.assert_equal(d, 0.0)


@numpyonly
def test_psi_dtw_2c():
    with util_numpy.test_uses_numpy() as np:
        x = np.arange(0, 20, .5)
        s1 = np.sin(x - 1)
        s2 = np.sin(x)
        d = dtw.distance_fast(s1, s2, psi=2, window=3)
        np.testing.assert_equal(d, 0.0)


@numpyonly
def test_twoleadecg_1(directory=None):
    with util_numpy.test_uses_numpy() as np:
        s1 = np.array([1.8896,-0.23712,-0.23712,-0.20134,-0.16556,-0.20134,-0.16556,-0.12978,-0.058224,0.013335,0.031225,0.10278,0.013335,-0.094004,-0.058224,-0.11189,-0.14767,-0.16556,-0.14767,-0.094004,-0.14767,-0.16556,-0.16556,-0.21923,-0.21923,-0.25501,-0.20134,-0.20134,-0.18345,-0.23712,-0.20134,-0.23712,-0.12978,-0.11189,-0.46969,-1.2747,-2.3481,-2.8133,-2.7775,-2.5986,-2.3839,-2.0082,-1.8651,-1.6146,-1.3463,-1.1495,-0.88115,-0.55914,-0.34446,-0.16556,-0.0045548,0.2459,0.53214,0.65737,0.71104,0.74682,0.76471,0.76471,0.80049,0.81838,0.87204,0.88993,0.97938,0.97938,1.0152,1.0867,1.1583,1.1762,1.212,1.2656,1.2656,1.2477,1.2656,1.1762,1.0867,0.99727,0.88993,0.74682,0.63948,0.58581,0.47847,0.38902])
        s2 = np.array([1,0.93163,0.094486,0.094486,0.038006,0.080366,0.080366,0.052126,0.080366,0.12273,0.22157,0.29217,0.41925,0.48985,0.39101,0.39101,0.30629,0.24981,0.19333,0.080366,-0.0043544,-0.018474,-0.089075,-0.11731,-0.14555,-0.17379,-0.21615,-0.27263,-0.20203,-0.315,-0.25851,-0.17379,-0.28675,-0.24439,0.16509,-0.11731,-1.0069,-1.9812,-2.4895,-2.786,-2.9272,-2.4612,-2.0518,-1.8964,-1.8258,-1.7411,-1.6705,-1.2893,-0.99276,-0.65388,-0.37148,-0.30087,-0.046714,0.30629,0.53221,0.65929,0.65929,0.72989,0.74401,0.87109,0.89933,0.95581,0.96993,1.0546,1.1394,1.2523,1.2523,1.2947,1.3088,1.3512,1.2806,1.2806,1.1394,1.097,0.89933,0.72989,0.67341,0.54633,0.37689,0.23569,0.10861,0.080366,-0.074955])
        d, paths = dtw.warping_paths(s1, s2, psi=2, window=5)
        path = dtw.warping_path(s1, s2, psi=2)
        if directory:
            dtwvis.plot_warping(s1, s2, path, filename=str(directory / "warping.png"))
            path = dtw.best_path(paths)
            dtwvis.plot_warpingpaths(s1, s2, paths, path, filename=str(directory / "warpingpaths.png"))


if __name__ == "__main__":
    with util_numpy.test_uses_numpy() as np:
        np.set_printoptions(precision=2, linewidth=120)
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    # test_normalize()
    # test_normalize2()
    test_psi_dtw_1a()
    # test_psi_dtw_1b()
    # test_psi_dtw_1c()
    # test_psi_dtw_2a()
    # test_psi_dtw_2b()
    # test_twoleadecg_1()




#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""
 python3 setup.py build_ext --inplace
"""
from setuptools import setup, Command
from setuptools.extension import Extension
from setuptools.command.test import test as TestCommand
from setuptools.command.sdist import sdist as SDistCommand
from setuptools.command.build_ext import build_ext as BuildExtCommand
from setuptools.command.install import install
from setuptools import Distribution
import platform
import os
import sys
import re
import subprocess as sp
from pathlib import Path

try:
    import numpy
    np_include_dirs = [numpy.get_include()]
except ImportError:
    numpy = None
    np_include_dirs = []

try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None

here = os.path.abspath(os.path.dirname(__file__))
dtaidistancec_path = Path('dtaidistance') / 'lib' / 'DTAIDistanceC' / 'DTAIDistanceC'

c_args = {
    # Xpreprocessor is required for the built-in CLANG on macos, but other
    # installations of LLVM don't seem to be bothered by it (although it's
    # not required.
    'unix': ['-Xpreprocessor', '-fopenmp',
             '-I'+str(dtaidistancec_path)],
    'msvc': ['/openmp', '/Ox', '/fp:fast', '/favor:INTEL64', '/Og',
             '/I'+str(dtaidistancec_path)],
    'mingw32': ['-fopenmp', '-O3', '-ffast-math', '-march=native',
                '-I'+str(dtaidistancec_path)]
}
l_args = {
    'unix': ['-fopenmp'],
    'msvc': [],
    'mingw32': ['-fopenmp']
}


class PyTest(TestCommand):
    description = "Run tests"
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]
    pytest_args = []
    test_args = []

    def initialize_options(self):
        self.pytest_args = ['--ignore=venv']
        try:
            import pytest_benchmark
            self.pytest_args += ['--benchmark-skip']
        except ImportError:
            print("No benchmark library, ignore benchmarks")
            self.pytest_args += ['--ignore', str(Path('tests') / 'test_benchmark.py')]

    def finalize_options(self):
        pass

    def run_tests(self):
        import pytest
        sys.path.append('.')
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


class MyDistribution(Distribution):
    global_options = Distribution.global_options + [
        ('noopenmp', None, 'Disable compilation with openmp')
    ]

    def __init__(self, attrs=None):
        self.noopenmp = 0
        super().__init__(attrs)


class MyInstallCommand(install):
    pass

    # def initialize_options(self):
    #     install.initialize_options(self)

    # def finalize_options(self):
    #     install.finalize_options(self)

    # def run(self):
    #     install.run(self)


def set_custom_envvars_for_homebrew():
    """Update environment variables automatically for Homebrew if CC is not set"""
    # DEPRECATED. OpenMP is now supported through -Xpreprocessor
    # if platform.system() == 'Darwin' and "CC" not in os.environ:
    #     print("Set custom environment variables for Homebrew Clang because CC is not set")
    #     cppflags = []
    #     if "CPPFLAGS" in os.environ:
    #         cppflags.append(os.environ["CPPFLAGS"])
    #     cflags = []
    #     if "CFLAGS" in os.environ:
    #         cflags.append(os.environ["CFLAGS"])
    #     ldflags = []
    #     if "LDFLAGS" in os.environ:
    #         ldflags.append(os.environ["LDFLAGS"])
    #     if os.path.exists("/usr/local/opt/llvm/bin/clang"):
    #         # We have a recent version of LLVM that probably supports openmp to compile parallel C code (installed using
    #         # `brew install llvm`).
    #         os.environ["CC"] = "/usr/local/opt/llvm/bin/clang"
    #         print("CC={}".format(os.environ["CC"]))
    #         ldflags += ["-L/usr/local/opt/llvm/lib"]
    #         cppflags += ["-I/usr/local/opt/llvm/include"]
    #         cflags += ["-I/usr/local/opt/llvm/include"]
    #         try:
    #             mac_ver = [int(nb) for nb in platform.mac_ver()[0].split(".")]
    #             if mac_ver[0] == 10 and mac_ver[1] >= 14:
    #                 # From Mojave on, the header files are part of Xcode.app
    #                 incpath = '-I/Applications/Xcode.app/Contents/Developer/Platforms/' + \
    #                           'MacOSX.platform/Developer/SDKs/MacOSX.sdk/usr/include'
    #                 cppflags += [incpath]
    #                 cflags += [incpath]
    #         except Exception as exc:
    #             print("Failed to check version")
    #             print(exc)
    #     else:
    #         # The default clang in XCode is compatible with OpenMP when using -Xpreprocessor
    #         pass
    #
    #     if len(cppflags) > 0:
    #         os.environ["CPPFLAGS"] = " ".join(cppflags)
    #         print("CPPFLAGS={}".format(os.environ["CPPFLAGS"]))
    #     if len(cflags) > 0:
    #         os.environ["CFLAGS"] = " ".join(cflags)
    #         print("CFLAGS={}".format(os.environ["CFLAGS"]))
    #     if len(ldflags) > 0:
    #         os.environ["LDFLAGS"] = " ".join(ldflags)
    #         print("LDFLAGS={}".format(os.environ["LDFLAGS"]))
    # else:
    #     print("Using the following environment variables:")
    #     print("CC={}".format(os.environ.get("CC", "")))
    #     print("CPPFLAGS={}".format(os.environ.get("CPPFLAGS", "")))
    #     print("CFLAGS={}".format(os.environ.get("CPLAGS", "")))
    #     print("LDFLAGS={}".format(os.environ.get("LDFLAGS", "")))


class MyBuildExtCommand(BuildExtCommand):

    def build_extensions(self):
        c = self.compiler.compiler_type
        print("Compiler type: {}".format(c))
        print("--noopenmp: {}".format(self.distribution.noopenmp))
        if self.distribution.noopenmp == 0:
            try:
                check_result = check_openmp(self.compiler.compiler[0])
            except Exception as exc:
                print("WARNING: Cannot check for OpenMP, assuming to be available")
                print(exc)
                check_result = True  # Assume to be present by default
            if not check_result:
                print("WARNING: OpenMP is not available, disabling OpenMP (no parallel computing in C)")
                self.distribution.noopenmp = 1
        if c in c_args:
            if self.distribution.noopenmp == 1:
                args = [arg for arg in c_args[c] if "openmp" not in arg]
            else:
                args = c_args[c]
            for e in self.extensions:
                e.extra_compile_args = args
        else:
            print("Unknown compiler type: {}".format(c))
        if c in l_args:
            if self.distribution.noopenmp == 1:
                args = [arg for arg in l_args[c] if "openmp" not in arg]
            else:
                args = l_args[c]
            for e in self.extensions:
                e.extra_link_args = args
        if numpy is None:
            self.extensions = [arg for arg in self.extensions if "numpy" not in str(arg)]
        print(f'All extensions:')
        print(self.extensions)
        BuildExtCommand.build_extensions(self)

    def initialize_options(self):
        set_custom_envvars_for_homebrew()
        super().initialize_options()

    # def finalize_options(self):
    #     super().finalize_options()

    # def run(self):
    #     super().run()


class MyBuildExtInPlaceCommand(MyBuildExtCommand):
    def initialize_options(self):
        super().initialize_options()
        self.inplace = True


def check_openmp(cc_bin):
    """Check if OpenMP is available"""
    print("Checking for OpenMP availability")
    cc_binname = os.path.basename(cc_bin)
    args = None
    kwargs = None
    if "clang" in cc_binname or "cc" in cc_binname:
        args = [[str(cc_bin), "-dM", "-E", "-fopenmp", "-"]]
        kwargs = {"stdout": sp.PIPE, "input": '', "encoding": 'ascii'}
        print(" ".join(args[0]) + " ".join(str(k) + "=" + str(v) for k, v in kwargs.items()))
    if args is not None:
        try:
            p = sp.run(*args, **kwargs)
            defs = p.stdout.splitlines()
            for curdef in defs:
                if "_OPENMP" in curdef:
                    print("... found OpenMP")
                    return True
        except Exception:
            print("... no OpenMP")
            return False
    else:
        print("... do not know how to check for OpenMP (unknown CC), assume to be available")
        return True
    return False


# Set up extension
extensions = []
if cythonize is not None:
    extensions.append(
        Extension(
            "dtaidistance.dtw_cc",
            [str(Path("dtaidistance") / "dtw_cc.pyx"), str(Path("dtaidistance") / "dtw_cc.pxd"),
             str(dtaidistancec_path / "dtw.c")],
            include_dirs=[],
            extra_compile_args=[],
            extra_link_args=[]))
    extensions.append(
        Extension(
            "dtaidistance.ed_cc",
            [str(Path("dtaidistance") / "ed_cc.pyx"),
             str(dtaidistancec_path / "ed.c")],
            include_dirs=[],
            extra_compile_args=[],
            extra_link_args=[]))
    extensions.append(
        Extension(
            "dtaidistance.dtw_cc_omp",
            [str(Path("dtaidistance") / "dtw_cc_omp.pyx"),
             str(dtaidistancec_path / "dtw_openmp.c")],
            include_dirs=[],
            extra_compile_args=[],
            extra_link_args=[]))

    if numpy is not None:
        extensions.append(
            Extension(
                "dtaidistance.dtw_cc_numpy", [str(Path("dtaidistance") / "util_numpy_cc.pyx")],
                include_dirs=np_include_dirs,
                extra_compile_args=[],
                extra_link_args=[]))
    else:
        print("WARNING: Numpy was not found, preparing a version without Numpy support.")

    ext_modules = cythonize(extensions)
else:
    print("WARNING: Cython was not found, preparing a pure Python version.")
    ext_modules = []

install_requires = ['cython']
tests_require = ['pytest']

# Check version number
with open('dtaidistance/__init__.py', 'r', encoding='utf-8') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)
if not version:
    raise RuntimeError('Cannot find version information')

# Set up readme file
readme_path = os.path.join(here, 'README.md')
if os.path.exists(readme_path):
    with open(readme_path, 'r', encoding='utf-8') as f:
        long_description = f.read()
else:
    long_description = ""
long_description_content_type = "text/markdown"

# Create setup
setup(
    name='dtaidistance',
    version=version,
    description='Distance measures for time series',
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    author='Wannes Meert',
    author_email='wannes.meert@cs.kuleuven.be',
    url='https://dtai.cs.kuleuven.be',
    project_urls={
        'DTAIDistance documentation': 'http://dtaidistance.readthedocs.io/en/latest/',
        'DTAIDistance source': 'https://github.com/wannesm/dtaidistance'
    },
    packages=["dtaidistance"],
    python_requires='>=3.5',
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        'vis': ['matplotlib'],
        'numpy': ['numpy']
    },
    include_package_data=True,
    package_data={
        'dtaidistance': ['*.pyx', '*.pxd', '*.c', '*.h'],
    },
    distclass=MyDistribution,
    cmdclass={
        'test': PyTest,
        'buildinplace': MyBuildExtInPlaceCommand,
        'build_ext': MyBuildExtCommand,
        'install': MyInstallCommand
    },
    license='Apache 2.0',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3'
    ],
    keywords='dtw',
    ext_modules=ext_modules
)





class NumpyException(Exception):
    def __init__(self, message):
        super().__init__(message)


class MatplotlibException(Exception):
    def __init__(self, message):
        super().__init__(message)


class ScipyException(Exception):
    def __init__(self, message):
        super().__init__(message)
